-- Ein einfacher monadischer Parser mit unbeschraenktem Backtracking.
--
-- In letzterem Punkt unterscheidet er sich fundamental von Parsec
-- (Text.ParserCombinators.Parsec)

module Parser (..) where

import Control.Monad
import Data.Char

-- Ein Parser ist eine Funktion, die eine Eingabezeichenkette auf
-- ein Ergebnis und eine Restzeichenkette abbildet.
data Parser a = Parser { runParser :: String -> [(a, String)] }

instance Functor Parser where
    fmap f g = g >>= return . f

instance Monad Parser where
    return a = Parser $ \s -> [(a, s)]

    x >>= f = Parser $ \s -> concatMap (\(a, s') -> runParser (f a) s')
                                       (runParser x s)

instance MonadPlus Parser where
    mzero = Parser $ \s -> []
    a `mplus` b = Parser $ \s -> runParser a s ++ runParser b s

-- Alternative. (Alias fuer mplus)
(<|>) :: Parser a -> Parser a -> Parser a
(<|>) = mplus

-- Gibt den verbleibenden Eingabestring zurueck
lookahead :: Parser String
lookahead = Parser $ \s -> [(s, s)]

-- Passt auf Ende des Eingabestrings
eof :: Parser ()
eof = do
    s <- lookahead
    guard (null s)

-- Passt auf Zeichen, die das gegebene Praedikat erfuellen.
satisfy :: (Char -> Bool) -> Parser Char
satisfy p = Parser $ \s -> case s of
    (x:xs) | p x -> [(x, xs)]
    _            -> []

-- Passt auf das uebergebene Zeichen
char :: Char -> Parser Char
char c = satisfy (c ==)

-- Passt auf Leerzeichen
space :: Parser Char
space = satisfy isSpace

-- Passt auf Ziffern
digit :: Parser Char
digit = satisfy isDigit

-- Passt auf die leere Zeichekette.
empty :: Parser ()
empty = return ()

-- Passt auf beliebig viele Wiederholungen des uebergebenen Parsers
many :: Parser a -> Parser [a]
many p = empty >> return [] `mplus` many1 p

-- Passt auf eine oder mehr Wiederholungen des uebergebenen Parsers
many1 :: Parser a -> Parser [a]
many1 p = do
    a <- p
    as <- many p
    return (a:as)

data ParseResult a = Unique a | Ambiguous a [a] | ParseError
    deriving Show

runParser' :: Parser a -> String -> ParseResult a
runParser' p s = case map fst (runParser (do x <- p; eof; return x) s) of
    []   -> ParseError
    x:[] -> Unique x
    x:xs -> Ambiguous x xs
