{-# LANGUAGE FlexibleContexts, UndecidableInstances #-}

module Heap where

import Control.Monad.State
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set
import qualified Data.Set as Set

import Signature (Signature(..), ITrans, transData, syntacticEquiv)
import Poset
import Equiv


newtype Cell = Cell Int deriving (Eq, Ord, Show)
newtype Field = Field Int deriving (Eq, Ord, Show)

instance Enum Cell where
  fromEnum (Cell c) = c
  toEnum n = Cell n

instance Enum Field where
  fromEnum (Field f) = f
  toEnum n = Field n

data Signature s => Heap s = Heap
  { nextCellHeap :: Cell,
    nextFieldHeap :: Field,
    cellsHeap :: Map Cell (s Field),
    fieldsHeap :: Map Field Cell }

type HeapOp s = State (Heap s)

emptyHeap :: Signature s => Heap s
emptyHeap = Heap (Cell 0) (Field 0) Map.empty Map.empty

freshCell :: Signature s => HeapOp s Cell
freshCell = State doIt
  where doIt s = let c = nextCellHeap s
		 in (c, s { nextCellHeap = succ c })

freshField :: Signature s => HeapOp s Field
freshField = State doIt
  where doIt s = let f = nextFieldHeap s
		 in (f, s { nextFieldHeap = succ f })

setCell :: Signature s => Cell -> s Field -> HeapOp s ()
setCell x y = modify doIt
  where doIt s = let cs = cellsHeap s
		 in s { cellsHeap = Map.insert x y cs }

setField :: Signature s => Field -> Cell -> HeapOp s ()
setField v x = modify doIt
  where doIt s = let fs = fieldsHeap s
		 in s { fieldsHeap = Map.insert v x fs }

getCell :: Signature s => Cell -> HeapOp s (s Field)
getCell x = gets doIt
  where doIt h = Map.findWithDefault undefined x (cellsHeap h)

getField :: Signature s => Field -> HeapOp s Cell
getField v = gets doIt
  where doIt h = Map.findWithDefault undefined v (fieldsHeap h)

getCellFields :: Signature s => Cell -> HeapOp s (s Cell)
getCellFields x = getCell x >>= fmapM getField


freshFields :: Signature s => s a -> HeapOp s (s Field)
freshFields = fmapM (const freshField)

newCell :: Signature s => s a -> HeapOp s (Cell, ITrans Field a)
newCell s =
  do
    x <- freshCell
    t <- freshFields s
    setCell x t
    return (x, transData t s)

consCell :: Signature s => s Cell -> HeapOp s Cell
consCell s =
    do x <- freshCell
       t <- freshFields s
       setCell x t
       fzipWithM_ setField t s
       return x


carrierHeap :: Signature s => Heap s -> Set Cell
carrierHeap h = Set.fromList (Map.keys (cellsHeap h))

opnHeap :: Signature s => Heap s -> Cell -> s Cell
opnHeap h =
    let opn1 x = Map.findWithDefault undefined x (cellsHeap h)
	opn2 x = Map.findWithDefault undefined x (fieldsHeap h)
     in fmap opn2 . opn1

heapCellEquiv :: Signature s => Heap s -> Equiv Cell
heapCellEquiv h = kernelEquiv (opnHeap h) syntacticEquiv

heapCotermEquiv :: Signature s => Equiv (Heap s, Cell)
heapCotermEquiv = kernelEquiv (uncurry opnHeap) syntacticEquiv

heapPoset :: Signature s => Poset (Heap s)
heapPoset = Poset (<=) emptyHeap    
    where h1 <= h2 = posetLE cellsPoset (dataHeap h1) (dataHeap h2)
	  cellsPoset = genericFlatMapPoset syntacticEquiv
	  dataHeap h = (carrierHeap h, opnHeap h)
