module InterpretVM where

import Program
import VM
import Heap
import Signature
import Value
import Ana (Interp)
import Poset
import Memory

import Control.Monad
import Control.Monad.State
import Data.FiniteMap
import Data.List
import Data.Maybe

data ProgramInterp a = FunInterp FunctionId Int [a]
		     | IdInterp a

instance Functor ProgramInterp where
    fmap f (FunInterp fid i xs) = FunInterp fid i (map f xs)
    fmap f (IdInterp x) = IdInterp (f x)

instance Signature ProgramInterp where
    FunInterp fid1 i1 xs1 ==* FunInterp fid2 i2 xs2 =
	fid1 == fid2 && i1 == i2 && xs1 == xs2
    IdInterp x1           ==* IdInterp x2           = x1 == x2
    fmapM f (FunInterp fid i xs) = (return . FunInterp fid i) =<< mapM f xs
    fmapM f (IdInterp x)         = return . IdInterp =<< f x
    ffold f e (FunInterp fid i xs) = foldl f e xs
    ffold f e (IdInterp x)         = f e x
    fzipWithM_ f (FunInterp fid1 i1 xs1) (FunInterp fid2 i2 xs2)
	| fid1 == fid2 && i1 == i2 = zipWithM_ f xs1 xs2
    fzipWithM_ f (IdInterp x1)           (IdInterp x2) = f x1 x2


interpret :: Program -> Interp Univ ProgramInterp Univ
interpret p (IdInterp x) =
    do c <- getCellFields x
       return (fmap IdInterp c)
interpret p (FunInterp fid i cs) =
    let f = lookupWithDefaultFM (functions p) undefined fid
	ins = map Plain cs
	root :: Univ Field
	(root, h0) = probeRoot f ins i
	flds = threadData root
	(vals, hs) = unzip (map (probeField f ins) flds)
	vcs = map (fmap fromPlain) vals
	td :: Field :=>: ProgramInterp Cell
	td = zip flds vcs
	h1 = fromJust (maxPoset heapPoset (h0 : hs))
     in State (const (fmap (fromJust . flip lookup td) root, h1))

data TrapOut = TrapOut (Univ Field) (Heap Univ)
instance Probe Value TrapOut where
    setValue r0 r v =
	case v
	of (Plain c) | r0 == r -> do d <- liftStateHeap (getCell c)
				     trapWith (TrapOut d . memHeap)
	   _ -> doSetValue r v

data TrapSet = TrapSet (ProgramInterp Value) (Heap Univ)
instance Probe Value TrapSet where
    setValue r0 r v = if r0 == r then trapWith (TrapSet (IdInterp v) . memHeap)
		      else doSetValue r v
    evalFunction r0 f (params, n) =
	let (ins, outs, diffs) = decomposeParams f params
	    i = elemIndex r0 outs
	 in case i
	    of Just i' -> trapWith (TrapSet (FunInterp (functionId f) i' ins) . memHeap)
	       _       -> doEvalFunction r0 f (params, n)

probeRoot :: Function -> [Value] -> Int -> (Univ Field, Heap Univ)
probeRoot f ins i =
    case run (completeEvalFunction (!! i) f ins)
    of Trapped (TrapOut y h) -> (y, h)

probeField :: Function -> [Value] -> Field -> (ProgramInterp Value, Heap Univ)
probeField f ins fld =
    case run (completeEvalFunction (const (HeapRef fld)) f ins)
    of Trapped (TrapSet x h) -> (x, h)


completeEvalFunction :: Probe p r => ([Value] -> p) -> Function -> [Value] -> Computation r ()
completeEvalFunction p f ins =
    let t = hnf (functionType f)
	n = length (domain t)
	m = length (codomain t)
	k = length (redomain t)
	in do outs <- replicateM m global
	      diffs <- replicateM k global
	      let outdiffs = outs ++ diffs
	       in doEvalFunction (p outdiffs) f (ins ++ outdiffs, 0)

run c = runStateT c emptyMemory


decomposeParams f params =
    let t = hnf (functionType f)
	n = length (domain t)
	m = length (codomain t)
	k = length (redomain t)
	(ins, outdiffs) = splitAt n params
	(outs, diffs) = splitAt m params
     in (ins, outs, diffs)
