module Poset where

import Data.Maybe
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set

import Equiv


data Poset a = Poset { posetLE :: a -> a -> Bool, posetMin :: a }

unitPoset :: Poset ()
unitPoset = Poset le ()
    where le ()() = True

pairPoset :: Poset a -> Poset b -> Poset (a, b)
pairPoset pa pb = Poset (<=) (posetMin pa, posetMin pb)
    where (a1, b1) <= (a2, b2) = posetLE pa a1 a2 && posetLE pb b1 b2

flatPoset :: Equiv a -> Poset (Maybe a)
flatPoset eqa = Poset eq Nothing
    where eq Nothing Nothing = True
	  eq (Just x1) (Just x2) = equivEQ eqa x1 x2
	  eq _ _ = False

liftMaybePoset :: Poset a -> Poset (Maybe a)
liftMaybePoset pa = Poset (<=) Nothing
    where Nothing <= _ = True
	  _ <= Nothing = False
	  Just x1 <= Just x2 = posetLE pa x1 x2

lexicalPoset :: Poset a -> Poset [a]
lexicalPoset pa = Poset (<=) []
    where [] <= _ = True
	  _ <= [] = False
	  (x1 : ys1) <= (x2 : ys2) = posetLE pa x1 x2 && (ys1 <= ys2)

prefixPoset :: Equiv a -> Poset [a]
prefixPoset (Equiv eq) = Poset isPrefixOfModulo []
    where isPrefixOfModulo [] _ = True
	  isPrefixOfModulo _ [] = False
	  isPrefixOfModulo (x1 : ys1) (x2 : ys2) =
	      eq x1 x2 && isPrefixOfModulo ys1 ys2

flatMapPoset :: Ord a => Equiv b -> Poset (Map a b)
flatMapPoset eqb = Poset fmLE Map.empty
    where fmLE m1 m2 = let compat k = equivEQ eqmb (Map.lookup k m1) (Map.lookup k m2)
			in all compat (Map.keys m1)
	  eqmb = liftMaybeEquiv eqb

genericFlatMapPoset :: Ord a => Equiv b -> Poset (Set a, a -> b)
genericFlatMapPoset (Equiv eq) = Poset gfmLE (Set.empty, undefined)
    where gfmLE (d1, f1) (d2, f2) =
	      let compat x = x `Set.member` d2 && eq (f1 x) (f2 x)
		  in all compat (Set.toList d1)

liftMapPoset :: Ord a => Poset b -> Poset (Map a b)
liftMapPoset pb = Poset fmLE Map.empty
    where fmLE m1 m2 = let compat k = posetLE mpb (Map.lookup k m1) (Map.lookup k m2)
			in all compat (Map.keys m1)
	  mpb = liftMaybePoset pb

subsetPoset :: Ord a => Poset (Set a)
subsetPoset = Poset ssLE Set.empty
    where ssLE s1 s2 = Set.union s1 s2 == s2

minPoset :: Poset a -> [a] -> Maybe a
maxPoset :: Poset a -> [a] -> Maybe a

minPoset pa [] = Nothing
minPoset pa (x : ys) = aux x ys
    where aux x [] = Just x
	  aux x (y : zs) | posetLE pa x y = aux x zs
			 | posetLE pa y x = aux y zs
			 | otherwise = Nothing

maxPoset pa [] = Nothing
maxPoset pa (x : ys) = aux x ys
    where aux x [] = Just x
	  aux x (y : zs) | posetLE pa y x = aux x zs
			 | posetLE pa x y = aux y zs
			 | otherwise = Nothing
