module Program where

import Data.List
import Data.Set
import Data.Maybe
import Data.FiniteMap

data Program = Program
    { types     :: FiniteMap TypeId     Type,
      constants :: FiniteMap ConstantId Constant,
      functions :: FiniteMap FunctionId Function
    }

emptyProgram = Program emptyFM emptyFM emptyFM

type Label = String
type Family = FiniteMap Label
type TypeId = String
type TSubst = FiniteMap TypeId Type

data Type = TVar TypeId | TConst TypeId
	  | Product (Family Type)
	  | Coproduct (Family (Maybe Type))
	  | Exponent
	    { domain :: [Type],
	      codomain :: [Type],
	      redomain :: [Type]
	    }
	  | Instantiation TSubst Type

whnf :: Type -> Type
whnf (Instantiation s t) = inst s t
whnf t = t

inst :: TSubst -> Type -> Type
inst s t =
    case t of
    TVar a             -> lookupWithDefaultFM s t a
    TConst c           -> TConst c
    Product ts         -> Product (inst' s ts)
			  where inst' = mapFM . const . Instantiation
    Coproduct ts       -> Coproduct (inst' s ts)
			  where inst' = mapFM . const . fmap . Instantiation
    Exponent ts us vs  -> Exponent (inst' s ts) (inst' s us) (inst' s vs)
	                  where inst' = map . Instantiation
    Instantiation s' t -> inst s (inst s' t)

data RType = ValueType Type
	   | RefType !RType

type FunctionId = String
data Function = Function
    { functionId :: FunctionId,
      functionType :: Type, -- must be Exponent
      principalBody :: Body,
      recycleBody :: Body
    }
type Body = Maybe Block

instance Show Function where
    show f = show (functionId f)

instance Eq Function where
    f1 == f2 = functionId f1 == functionId f2

inParams, outParams, diffParams :: Function -> [Type]
inParams = domain . whnf . functionType
outParams = codomain . whnf . functionType
diffParams = redomain . whnf . functionType

data Block = Block [RType] [Statement]

type ConstantId = String
data Constant = Constant
    { constantId :: ConstantId,
      constantType :: Type,
      initializer :: Function
    }

instance Eq Constant where
    c1 == c2 = constantId c1 == constantId c2
instance Ord Constant where
    compare c1 c2 = compare (constantId c1) (constantId c2)

data Statement = SBlock Block
	       | Pop | Dup | Swap
	       | Return | Halt
	       | Param Int | Var Int
	       | Insitu RType
	       | Get | Set
	       | Tuple Type
	       | Cotuple Label Type
	       | Field Label
	       | Body Label
	       | Yield Constant
	       | Call Function
	       | Loop | Dito
	       | Curry Function
	       | Apply Int
	       | Eval
	       | Switch (FiniteMap Label Block)
