{-# LANGUAGE TypeOperators, FlexibleContexts #-}

module Signature (Signature(..), syntacticEquiv,
		  Identity(..), Const(..),
		  (:*:)(..), tupling,
		  (:+:)(..), cotupling,
		  localData,
		  ITrans, transData, transCover, threadData, succData) where

import Control.Monad.Identity
import Control.Monad.State
import Data.Set (Set, insert, empty)


import Equiv


class Functor s => Signature s where
  (==*) :: Eq x => s x -> s x -> Bool
  fmapM :: Monad m => (a -> m b) -> s a -> m (s b)
  ffold :: (a -> b -> a) -> a -> s b -> a 
  fzipWithM_ :: Monad m => (a -> b -> m ()) -> s a -> s b -> m ()

syntacticEquiv :: (Signature s, Eq x) => Equiv (s x)
syntacticEquiv = Equiv (==*)


instance Signature Identity where
  (Identity x) ==* (Identity x') = x == x'
  fmapM f (Identity x) = liftM Identity (f x)
  ffold f e (Identity x) = f e x
  fzipWithM_ f (Identity x) (Identity y) = f x y


newtype Const a x = Const a
instance Functor (Const a) where
  fmap f (Const a) = (Const a)
instance Eq a => Signature (Const a) where
  (Const x) ==* (Const x') = x == x'
  fmapM f (Const x) = return (Const x)
  ffold f e (Const x) = e
  fzipWithM_ f (Const x) (Const x') | x == x' = return ()

data (:*:) f g x = (:*:) { project1 :: f x, project2 :: g x }
instance (Functor f, Functor g) => Functor (f :*: g) where
  fmap f (x :*: y) = fmap f x :*: fmap f y
instance (Signature f, Signature g) => Signature (f :*: g) where
  (x :*: y) ==* (x' :*: y') = (x ==* x') && (y ==* y')
  fmapM f (x :*: y) = {-do x' <- fmapM f x
                         y' <- fmapM f y
                         return (x' :*: y')-}
                      liftM2 (:*:) (fmapM f x) (fmapM f y)
  ffold f e (x :*: y) = ffold f (ffold f e x) y
  fzipWithM_ f (x :*: y) (x' :*: y') =
    fzipWithM_ f x x' >> fzipWithM_ f y y'

tupling :: (a -> f b) -> (a -> g b) -> a -> (f :*: g) b
tupling f g x = f x :*: g x

fproduct :: (f a -> f b) -> (g a -> g b) -> (f :*: g) a -> (f :*: g) b
fproduct f g = tupling (f . project1) (g . project2)


data (:+:) f g x = Inject1 { part1 :: f x } | Inject2 { part2 :: g x }
instance (Functor f, Functor g) => Functor (f :+: g) where
  fmap f (Inject1 x) = Inject1 (fmap f x)
  fmap f (Inject2 x) = Inject2 (fmap f x)
instance (Signature f, Signature g) => Signature (f :+: g) where
  Inject1 x ==* Inject1 x' = x ==* x'
  Inject2 x ==* Inject2 x' = x ==* x'
  _ ==* _ = False
  fmapM f (Inject1 x) = fmapM f x >>= return . Inject1
  fmapM f (Inject2 x) = fmapM f x >>= return . Inject2
  ffold f e (Inject1 x) = ffold f e x
  ffold f e (Inject2 x) = ffold f e x
  fzipWithM_ f (Inject1 x) (Inject1 x') = fzipWithM_ f x x'
  fzipWithM_ f (Inject2 x) (Inject2 x') = fzipWithM_ f x x'

cotupling :: (f a -> b) -> (g a -> b) -> (f :+: g) a -> b
cotupling f g (Inject1 x) = f x
cotupling f g (Inject2 x) = g x

fcoproduct :: (f a -> f b) -> (g a -> g b) -> (f :+: g) a -> (f :+: g) b
fcoproduct f g = cotupling (Inject1 . f) (Inject2 . g)


newtype List f x = List [f x]
instance Functor f => Functor (List f) where
  fmap f (List xs) = List (map (fmap f) xs)
instance Signature f => Signature (List f) where
  List [] ==* List [] = True
  List (x : ys) ==* List (x' : ys') = (x ==* x') && (List ys ==* List ys')
  List _ ==* List _ = False
  fmapM f (List xs) = mapM (fmapM f) xs >>= return . List
  ffold f e (List xs) = foldl (ffold f) e xs
  fzipWithM_ f (List xs) (List xs') = zipWithM_ (fzipWithM_ f) xs xs'


localData :: Signature s => s a -> s ()
localData = fmap (const ())

compatible :: Signature s => s a -> s a -> Bool
compatible x x' = localData x ==* localData x'


type ITrans a b = [(a, b)]

transData :: Signature s => s a -> s b -> ITrans a b
{-transData x x' = reverse (execState (fzipWithM_ update x x') [])
  where update :: a -> b -> State (ITrans a b) ()
	update y y' = modify ((y, y') :)-}
transData x x' = execState (fzipWithM_ update x x') []
  where update :: a -> b -> State (ITrans a b) ()
	update y y' = modify (++ [(y, y')])

transCover :: Monad m => (a -> b -> m ()) -> ITrans a b -> m ()
transCover = mapM_ . uncurry



threadData :: Signature s => s a -> [a]
threadData = reverse . ffold (flip (:)) []

neighbours :: Signature s => s a -> [a]
neighbours = ffold (\ xs y -> (xs ++ [y])) []

succData :: (Signature s, Ord a) => s a -> Set a
succData = ffold (flip insert) empty



class Signature s => PolySignature s where
  fzip :: s a -> s b -> s (a, b)

instance PolySignature Identity where
  fzip (Identity x) (Identity y) = Identity (x, y)

instance Eq a => PolySignature (Const a) where
  fzip (Const a) (Const a') | a == a' = Const a

instance (PolySignature f, PolySignature g) => PolySignature (f :*: g) where
  fzip (x :*: y) (x' :*: y') = fzip x x' :*: fzip y y'

instance (PolySignature f, PolySignature g) => PolySignature (f :+: g) where
  fzip (Inject1 x) (Inject1 x') = Inject1 (fzip x x')
  fzip (Inject2 x) (Inject2 x') = Inject2 (fzip x x')

instance PolySignature f => PolySignature (List f) where
  fzip (List xs) (List ys) = List (zipWith fzip xs ys)


{-
instance Show a => Show (Identity a) where
  show (Identity x) = show x

instance Show a => Show (Const a b) where
  show (Const c) = show c

instance (Show (f a), Show (g a)) => Show ((f :*: g) a) where
  show (x :*: y) = "(" ++ show x ++ " :*: " ++ show y ++ ")"

instance (Show (f a), Show (g a)) => Show ((f :+: g) a) where
  show (Inject1 x) = "Lt" ++ show x
  show (Inject2 y) = "Rt" ++ show y
-}
