module Stack where

import Control.Monad.State
import Data.FiniteMap
import Data.List

import Heap
import Program
import Value
import Equiv

data Stack = Stack 
    { stackFrames :: [Frame],
      stackNextSlot :: Slot,
      stackSlots :: FiniteMap Slot Value
    }

modifyFrames :: ([Frame] -> [Frame]) -> Stack -> Stack
modifyFrames f st = st { stackFrames = f (stackFrames st) }

modifyNextSlot :: (Slot -> Slot) -> Stack -> Stack
modifyNextSlot f st = st { stackNextSlot = f (stackNextSlot st) }

modifySlots :: (FiniteMap Slot Value -> FiniteMap Slot Value) -> Stack -> Stack
modifySlots f st = st { stackSlots = f (stackSlots st) }

emptyStack = Stack { stackFrames = [],
		     stackNextSlot = Slot 0,
		     stackSlots = emptyFM }

data Frame = Frame
    { frameFunction :: Function,
      params :: [Value],
      opStack :: [Slot],
      vars :: [Slot]
    }

initFrame f xs =
    Frame { frameFunction = f,
	    params = xs,
	    opStack = [],
	    vars = []
	  }
			
modifyOpStack :: ([Slot] -> [Slot]) -> Frame -> Frame
modifyOpStack f fr = fr { opStack = f (opStack fr) }

modifyVars :: ([Slot] -> [Slot]) -> Frame -> Frame
modifyVars f fr = fr { vars = f (vars fr) }

liftStateFrame :: State Frame a -> State Stack a
liftStateFrame op = State (\ s ->
			   let (y, f) = (runState op . head . stackFrames) s
			   in (y, modifyFrames ((f :) . tail) s))

consumeSlot :: State Frame Slot
consumeSlot = State (\ f ->
		     case opStack f
		     of x : ys -> (x, f { opStack = ys }))

produceSlot :: Slot -> State Frame ()
produceSlot sl = modify (modifyOpStack (sl :))

freshSlotInit :: Value -> State Stack Slot
freshSlotInit v = do
		  sl <- freshSlot
		  setSlot sl v
		  return sl

freshSlot :: State Stack Slot
freshSlot = State doIt
    where doIt s = let f = stackNextSlot s
    		   in  (f, s { stackNextSlot = succ f })

setSlot :: Slot -> Value -> State Stack ()
setSlot sl v = modify (modifySlots (\ m -> addToFM m sl v))

consume :: State Stack Value
consume = getSlot =<< liftStateFrame consumeSlot

produce :: Value -> State Stack ()
produce v = do sl <- freshSlotInit v
	       liftStateFrame (modify (modifyOpStack (sl :)))

getSlot :: Slot -> State Stack Value
getSlot sl = gets (lookup sl)
    where lookup sl st = lookupWithDefaultFM (stackSlots st) undefined sl

param, var :: Int -> State Stack Value
param i = liftStateFrame (gets ((!! i) . params))
var i = liftStateFrame (gets (StackRef . (!! i) . vars))

addVar :: State Stack ()
addVar = do sl <- freshSlot
	    liftStateFrame (modify (modifyVars (++ [sl])))

removeVar :: State Stack ()
removeVar = liftStateFrame (modify (modifyVars init))

insitu :: State Stack ()
insitu = do sl <- freshSlot
	    liftStateFrame (produceSlot sl)
	    setSlot sl (StackRef sl)

isInsituRef :: Slot -> State Stack Bool
isInsituRef sl =
    do r <- getSlot sl
       case r
         of StackRef r | sl == r -> return True
	    _                    -> return False

enter :: Function -> [Value] -> State Stack ()
enter f xs = modify (modifyFrames (initFrame f xs :))

leave :: [Value] -> State Stack ()
leave xs = do modify (modifyFrames tail)
	      liftStateFrame (mapM_ produceSlot (map fromStackRef xs))

currentFunction :: State Stack Function
currentFunction = gets (frameFunction . head . stackFrames)

cycleCheck :: State Stack (Maybe Frame)
cycleCheck =
    do fr0 : frs <- gets stackFrames
       return (find (equivEQ cycleCheckEquiv fr0) frs)

cycleCheckEquiv :: Equiv Frame
cycleCheckEquiv = Equiv (~=)
    where fr ~= fr' =
	      let f = frameFunction fr
		  f' = frameFunction fr'
		  n = length (inParams f)
		  ins = take n (params fr)
		  ins' = take n (params fr')
		  in f == f' && ins == ins'
