module VM where

import Data.FiniteMap (mapFM, lookupWithDefaultFM)
import Data.Maybe (isNothing)

import Heap (getField, setField, newCell, getCell, getCellFields,
	     consCell, Heap, emptyHeap)
import Value (Value(..))
import qualified Stack (consume, produce)
import Stack (Stack(..), Frame(..), getSlot, setSlot, emptyStack, param, var,
	      addVar, removeVar, insitu, isInsituRef, freshSlot,
	      currentFunction, enter, leave, cycleCheck)
import Memory (Univ(..), Memory(..))
import Program (Type(..), whnf, Function(..), Constant(..), Statement(..), Block(..), inParams, diffParams)

import Computation (Computation, liftStateStack, liftStateHeap, transferControl)
import qualified Computation (halt, return)

import Control.Monad.State
import Data.FiniteMap (FiniteMap)


execMain :: Function -> [Value] -> Computation [Value]
execMain f xs = 
    do ys <- replicateM (length (codomain (functionType f))) global
       execFunction f (xs ++ ys) 0
       mapM getValue ys

execStatement :: Statement -> Computation ()

execStatement Pop = do consume
		       return ()
execStatement Dup = do x <- consume
		       produce x
		       produce x
execStatement Swap = do x <- consume
			y <- consume
			produce x
			produce y

execStatement Get = consume >>= getValue >>= produce
execStatement Set = do v <- consume
		       r <- consume
		       setValue r v

execStatement (Param i) = liftStateStack (param i) >>= produce
execStatement (Var i) = liftStateStack (var i) >>= produce
execStatement (Insitu t) = liftStateStack Stack.insitu

execStatement (Yield c) = gets getConst >>= produce
    where getConst mem = lookupFM' (memConstants mem) (constantId c)

execStatement Halt = Computation.halt
execStatement Return = Computation.return

execStatement (Tuple t) = alloc (UTuple ts) >>= produce
    where Product ts = whnf t
execStatement (Cotuple l t) = alloc (UCotuple l b) >>= produce
    where Coproduct ts = whnf t
	  b = (lookupFM' ts l)

execStatement (Field l) =
    do Plain c <- consume
       UTuple m <- liftStateHeap (getCell c)
       produce (HeapRef (lookupFM' m l))
       
execStatement (Body l) =
    do Plain c <- consume
       UCotuple l' b <- liftStateHeap (getCell c)
       case b of Just b' | l == l' -> produce (HeapRef b')

execStatement (Switch cases) =
    do Plain c <- consume
       UCotuple l b <- liftStateHeap (getCell c)
       unpack b
       execBlock (lookupFM' cases l)
    where unpack Nothing = return ()
	  unpack (Just b) = do c' <- liftStateHeap (getField b)
			       produce (Plain c')

execStatement (SBlock b) = execBlock b

execStatement (Curry f) =
    do c <- liftStateHeap (consCell (UClosure f []))
       produce (Plain c)

execStatement (Apply n) =
    do Plain c <- consume
       UClosure f xFields <- liftStateHeap (getCell c)
       xCells <- liftStateHeap (mapM getField xFields)
       yValues <- replicateM n consume
       let yCells = fmap fromPlain yValues
       c' <- liftStateHeap (consCell (UClosure f (xCells ++ yCells)))
       produce (Plain c')

execStatement (Call f) =
    consumeArgs (whnf (functionType f)) >>= callFunction f

execStatement Eval =
    do Plain c <- consume
       UClosure f xCells <- liftStateHeap (getCellFields c)
       let xValues = map Plain xCells
       (yValues, n) <- consumeOutArgs (whnf (functionType f))
       callFunction f (xValues ++ yValues, n)

execStatement Loop =
    do f <- liftStateStack currentFunction
       execStatement (Call f)
       execStatement Return

execStatement Dito = dito

dito :: Computation ()
dito =
    do inner : _ <- liftStateStack (gets stackFrames)
       Just outer <- liftStateStack cycleCheck
       let f = frameFunction inner
           n = length (inParams f)
	   0 = length (diffParams f)
	   out = drop n . params
           copy a b = getValue a >>= setValue b
       zipWithM_ copy (out outer) (out inner)

alloc :: Univ a -> Computation Value
alloc s = do (c, _) <- liftStateHeap (newCell s)
	     return (Plain c)

consume = liftStateStack Stack.consume
produce = liftStateStack . Stack.produce

getValue (StackRef r) = liftStateStack (getSlot r)
getValue (HeapRef r) = liftStateHeap (fmap Plain (getField r))
setValue (StackRef r) v = liftStateStack (setSlot r v)
setValue (HeapRef r) (Plain c) = liftStateHeap (setField r c)
callFunction f (xs, k) = execFunction f xs k

execBlock :: Block -> Computation ()
execBlock (Block vs ss) =
    do let n = length vs
       liftStateStack (replicateM_ n addVar)
       mapM_ execStatement ss
       liftStateStack (replicateM_ n removeVar)

consumeArgs, consumeOutArgs :: Type -> Computation ([Value], Int)
consumeArgs t =
    do xs <- replicateM (length (domain t)) consume
       (ys, n) <- consumeOutArgs t
       return (xs ++ ys, n)

consumeOutArgs t =
    do (xs, m) <- consumeOutArgsInsitu (codomain t) 0
       (ys, n) <- consumeOutArgsInsitu (redomain t) m
       return (xs ++ ys, n)

consumeOutArgsInsitu :: [Type] -> Int -> Computation ([Value], Int)
consumeOutArgsInsitu [] n = return ([], n)
consumeOutArgsInsitu (t : ts) k = consume >>= check
    where check x =
	      case x
	      of StackRef r -> do i <- liftStateStack (Stack.isInsituRef r)
				  if i then consumeOutArgInsitu x
				     else consumeOutArg x
		 _ -> consumeOutArg x
	  consumeOutArg x | k == 0 = do (ys, n) <- consumeOutArgsInsitu ts 0
					return (x : ys, n)
	  consumeOutArgInsitu x = 
	    do (ys, n) <- consumeOutArgsInsitu ts (k + 1)
	       return (x : ys, n + 1)



execFunction :: Function -> [Value] -> Int -> Computation ()
execFunction f xs n =
    do liftStateStack (enter f xs)
       body <- chooseBody (principalBody f) (recycleBody f)
       transferControl (execBlock body)
       liftStateStack (leave (take n (reverse xs)))
    where chooseBody (Just b) Nothing = return b
	  chooseBody (Just b) (Just b') =
	      fmap (maybe b (const b')) (liftStateStack cycleCheck)



global :: Computation Value
global = fmap StackRef (liftStateStack Stack.freshSlot)


lookupFM' :: Ord a => FiniteMap a b -> a -> b
lookupFM' = flip lookupWithDefaultFM undefined
